;*********************************************************************
; Commodore VIC 20 Software Sprite Stack - MMX Edition
; last updated: 22-Sep-2011
; 2015 AF Added 3x2 and multicolor support for Pooyan and Popeye
; written by Robert Hurst <robert@hurst-ri.us>
; with contributions from:
; Matt Dawson <matsondawson@gmail.com>
;
; Modifed 2023 by thegg to allow assembly using CBM prg Studio
; Also added the API element SSSLOADBG 
;*********************************************************************


DEBUG_CHECKS = 0               


;*********************************************************************
; Software Sprite Stack INITIALIZATION
;
; MUST BE INVOKED ONCE BEFORE USING ANY OTHER SSS CALL
; Value in COLORCODE will be used to fill the color buffers.
;
SSSINIT

                ; SSS geometry
                LDA VIC+$02
                AND #$1F
                STA PLAYCOLS
                ASL                             ; CLC
                ADC #$04
                ASL
                ASL
                STA SSSCLIPX
                LDA VIC+$03
                AND #$7E
                LSR
                STA PLAYROWS
                ASL                             ; CLC
                ADC #$04
                ASL
                ASL
                STA SSSCLIPY
                ;
                LDA #$00
                STA CRSRROW
                STA CRSRCOL
                TAY
                TAX
@sss    TYA
                STA sss+1,X
                LDA CRSRCOL
                STA sss,X
                CLC
                ADC PLAYCOLS
                BCC @cc
                INY
@cc     STA CRSRCOL
                INX
                INX
                INC CRSRROW
                LDA CRSRROW
                CMP PLAYROWS
                BNE @sss
        ;
                ; SSS active / pending video page pointers
                LDA #>VICFRAME1
                STA SCRNPAGE
                LDA #>VICFRAME2
                STA PENDING
                LDA #$40
                STA NEWDIRT
                ;
                ; kernal init
                LDA #$01
                STA RVSFLAG             ; character reverse flag
                LDA #$80
                STA SHIFTMODE   ; locked
                LDA #$00
                STA SCROLLFLAG  ; disable
                ;
                ; SPRITE register init
                STA SPRITES
                ;
                ; fill VIC screen / color buffers
                LDA #6
                STA COLORCODE   
                LDA #SSSNULL
                JSR SSSCLEAR
                ;
                ; VIC register init
                LDA VIC+$02
                AND #$7F                ; if $80 enabled, +$0200 to base screen address
                STA VIC+$02
                LDA #UDCPOKE            ; Set origin of screen and character definition data
                STA VIC+$05
                ;
                ; FRAME register init
                LDY #$00
                STY FPS
                STY VSYNC
                STY VSYNC2
                STY VCOUNT
                JSR SSSFLIP

                RTS


;*********************************************************************
; Software Sprite Stack IRQ HANDLER
;
; PROGRAM USE OF THIS HANDLER IS OPTIONAL
;
; customize to your liking, i.e.,
; - change JMP $EABF here to continue to your custom IRQ handler; or
; - JMP SSSIRQ at the end of your IRQ handler.
;
SSSIRQ 
                LDA JIFFYL
                AND #%00111111
                BNE @cont
                LDX VCOUNT
                STX FPS
                STA VCOUNT
@cont 
                INC VSYNC2              ; frame skipped?
                LDA VSYNC
                BEQ @fskip              ; program is NOT waiting ...
                LDX #$00
                SEC
                SBC VSYNC2
                STA VSYNC               ; save result
                BCS @reset              ; program wants more than 1 screen refresh
                STX VSYNC               ; clear wait for vertical sync flag
@reset          STX VSYNC2              ; clear frame skip counter
@fskip 
               
                JMP $EABF


;*********************************************************************
; Software Sprite Stack CLEAR SCREEN
;
; Pass A with the character code to fill the PLAYFIELD buffer.
; Value in COLORCODE will be used to fill the PLAYCOLOR buffer.
;
; These changes will go into effect only after a call to SSSFLIP
;
SSSCLEAR 
                PHA                  ;++ save character fill code
                LDX #$00
                LDY #$00
                JSR SSSPLOT          ; home "cursor"
@reset          TYA
                STA DIRTYLINE,X
                LDA PLAYCOLS
                STA DIRTYLINE2,X
                INX
                CPX PLAYROWS
                BNE @reset
                ;
@cls            PLA                  ;--
                PHA                  ;++
                JSR SSSPRINT
                LDX CRSRCOL
                BNE @cls             ; loop until column wraps
                LDY CRSRROW
                BNE @cls             ; loop until row wraps, too
                PLA                  ;--
                RTS


;*********************************************************************
; Software Sprite Stack PLOT CURSOR IN FRAME BUFFER
;
; SSSPLOT uses PLAYFIELD frame buffer.
; SSSPLOTS uses PENDING frame buffer.
;
; Pass X/Y with coordinate to put cursor.
;
SSSPLOT 
                LDA #>PLAYFIELD
                STA SCRNLINE+1  ;$d2 upper map
                LDA #>PLAYCOLOR
                STA COLORLINE+1 ;$f4 upper colour map
                BNE SSSPLOTX
;
; PENDING frame cursor (for writing SPRITE chars)
SSSPLOTS 
                LDA PENDING
                STA SCRNLINE+1
                ORA #$84
                STA COLORLINE+1
;
SSSPLOTX 
                LDA #>PLAYCOLOR
                STA DIRTMAP+1
@x              CPX PLAYCOLS
                BMI @y
                LDX PLAYCOLS
                DEX
@y              CPY PLAYROWS
                BMI @ok
                LDY PLAYROWS
                DEY
@ok             STX CRSRCOL             ; maintain column offset to row
                STY CRSRROW             ; maintain row number
                TYA
                ASL
                TAY
                LDA sss+1,Y
                BEQ @top
                INC SCRNLINE+1
                INC COLORLINE+1
                INC DIRTMAP+1
@top            LDA sss,Y
                STA SCRNLINE
                STA COLORLINE
                STA DIRTMAP
                LDY CRSRROW
                RTS


;*********************************************************************
; Software Sprite Stack PRINT TO A FRAME BUFFER
;
; Like SSSPOKE, writes to current cursor, but also advances it to
; the right (with line/screen wrap) upon completion.
;
; All registers are preserved from this call, for common loop use.
;
SSSPRINT 
                STX XCOPY        ; save index registers
                STY YCOPY
                JSR SSSPOKE
                INC CRSRCOL     ; cursor right
                LDA CRSRCOL
                CMP PLAYCOLS    ; moved past right edge?
                BCC @fini         ; no, all done
                LDA #$00
                STA CRSRCOL     ; reset to 1st column
                INC CRSRROW     ; and advance down a row
                LDA CRSRROW
                CMP PLAYROWS    ; moved past bottom edge?
                BCS @toprow         ; yes, wrap back to top
                LDA SCRNLINE    ; no, re-calculate new row pointer
                CLC
                ADC PLAYCOLS
                BCC @cc
                INC SCRNLINE+1
                INC COLORLINE+1
@cc             STA SCRNLINE
                STA COLORLINE
@fini           LDY CRSRCOL
                LDA (COLORLINE),Y
                STA CRSRCOLOR
                LDA (SCRNLINE),Y
                STA CRSRCHAR
                LDX XCOPY               ; restore index registers
                LDY YCOPY
                RTS
@toprow 
                LDA #$00
                STA CRSRROW
                STA SCRNLINE
                STA COLORLINE
                DEC SCRNLINE+1
                DEC COLORLINE+1
                BNE @fini


;*********************************************************************
; Software Sprite Stack PRINT STRING FROM POINTER ON STACK
;
; Will print bytes, until a NULL, following the JSR call here.
; Carriage control and color codes are interpreted.
; Note that there is a limit of 256 bytes of data
;
SSSXYPRINTS 
                PLA
                STA VECTORFG
                PLA
                STA VECTORFG+1
                LDY #$01
                LDA (VECTORFG),Y
                PHA
                INY
                LDA (VECTORFG),Y
                TAY
                PLA
                TAX
                JSR SSSPLOT
                LDY #3
                JMP DOSSSPRINTS
                
SSSPRINTS 
                PLA
                STA VECTORFG
                PLA
                STA VECTORFG+1
                LDY #$01
DOSSSPRINTS 
@loop           LDA (VECTORFG),Y ;limits data length to 256
                BEQ @fini
                CMP #$0D         ; carriage return
                BNE @ctrl
                TYA
                PHA
                LDY CRSRROW
                INY
                LDX #0
                JSR SSSPLOT
                PLA
                TAY
                JMP @next
@ctrl           CMP #$F0
                BCC @cont               ; color code?
                AND #$0F                ; filter for 16-colors
                STA COLORCODE   ; 0=blk,1=wht,2=red,3=cyn,4=mag,5=grn,6=blu,7=yel
                JMP @next
@cont           JSR SSSPRINT
@next           INY
                BNE @loop
                INC VECTORFG+1
                BNE @loop
                ;
@fini           TYA
                CLC
                ADC VECTORFG
                BCC @cc
                INC VECTORFG+1
@cc             STA VECTORFG
                LDA VECTORFG+1
                PHA
                LDA VECTORFG
                PHA
                RTS


;*********************************************************************
; Software Sprite Stack READ FROM A FRAME BUFFER
;
; SSSPEEK reads from PLAYFIELD frame buffer.
; SSSPEEKS reads from PENDING frame buffer.
; SSSPEEKXY reads from PLAYFIELD frame buffer, using the sprite pixel
; coordinate system to determine X/Y cursor positioning.
;
; Pass X/Y with the coordinate to put cursor.
; CRSRCHAR and CRSRCOLOR are filled with values under cursor, with
; the former returned in Accumulator.
;
SSSPEEKS 
                JSR SSSPLOTS
                JMP SSSPEEKX
;
; pixel X/Y
SSSPEEKXY 
                CPX #$10
                BCC @fini               ; hidden by left border
                CPX SSSCLIPX
                BCS @fini               ; hidden by right border
                CPY #$10
                BCC @fini               ; above top border
                CPY SSSCLIPY
                BCC @ok                 ; below bottom border
@fini           RTS
@ok     
                 TYA
                 SEC
                 SBC #$10
                 TAY
                 LDA LSR3,Y  ; Y / 8
                 TAY
                ;
                 TXA
                 SEC
                 SBC #$10
                 TAX
                 LDA LSR3,X  ; X / 8
                 TAX
;
SSSPEEK 
                JSR SSSPLOT
;
SSSPEEKX 
                LDY CRSRCOL
                LDA (COLORLINE),Y
                STA CRSRCOLOR
                LDA (SCRNLINE),Y
                STA CRSRCHAR
                RTS


;*********************************************************************
; Software Sprite Stack WRITE TO A FRAME BUFFER
;
; Pass Accumulator with character code to write to current cursor.
; COLORCODE is used to fill same space with that value.
;
SSSPOKE 
                LDY CRSRCOL
                STA (SCRNLINE),Y
                LDA (DIRTMAP),Y
                LDX SCRNLINE+1
                CPX #>PLAYFIELD
                BPL @bg
                ORA NEWDIRT
                STA (DIRTMAP),Y ; update sprite's dirty bits only
ifdef SPRITEDEF5
                LDA (COLORLINE),Y
                BIT COLORCODE
                BMI @color      ; keep color in place
endif
                LDA COLORCODE   ; add color directly to map
                JMP @color
@bg             AND #%11000000  ; keep any sprite dirt
                ORA #%00100000  ; dirty this PLAYFIELD cell
                ORA COLORCODE   ; add color
@color          STA (COLORLINE),Y
                LDX CRSRROW
                LDY CRSRCOL
                LDA DIRTYLINE,X
                CMP CRSRCOL
                BCC @ok         ; is dirty seek lower than this write?
                STY DIRTYLINE,X ; start looking for dirty cells from here
@ok             TYA
                CMP DIRTYLINE2,X
                BCC @ok2         ; is dirty seek higher than this write?
                INY
                STY DIRTYLINE2,X ; stop looking for dirty cells after here
@ok2            RTS


;*********************************************************************
; Software Sprite Stack FLIP ACTIVE / PENDING FRAME BUFFERS
;
; Pass Y with the number of vertical sync counts to wait for, or zero
; to make changes visible immediately.
;
; Only the current cursor position is preserved when completed.
;
; If user is holding RUN/STOP key down, the current video frame is
; paused until it is released.
;
SSSFFLIP 
                LDA #$00        ; gameplay needs its action to move faster
                BIT VSYNC2
                BPL SSSFLIP2    ; do we need to drop a frame?
                STA VSYNC2
                RTS
                ;
SSSFLIP 
                LDA #$00
                STA VSYNC2
SSSFLIP2 
                LDA CRSRROW
                PHA                             ;++ save row
                LDA CRSRCOL
                PHA                             ;++ save column
                TYA
                PHA                             ;++ save Y (frame count)
                ;
                ; Phase I:
                ; - write any new PLAYFIELD updates to PENDING frame
                ; - erase any SPRITE characters from PENDING frame
                ;
                LDA #%00010000  ; clean TOPFIELD dirt only after this update
                JSR SSSCOMMIT
                ;
                ; Phase II:
                ; - render & write SPRITE characters to PENDING frame
                ;
                JSR SSSUPDATE
                ;
                ; PHASE III:
                ; - signal IRQ to flip video to PENDING frame
                ; - wait for the all clear
                ;
                PLA                     ;-- restore A (frame count)
                TAY
                INY                     ; account for a 'missed' frame
        ;       INY                     ; allow for another 'missed' frame
                CPY VSYNC2
                BCS @pace               ; is game loop & rendering ok?
                LDY #$80                ; nope, flag next call to SSSFFLIP
                STY VSYNC2              ; to skip rendering/flip altogether
                LDA #$00                ; and don't wait for this vsync either
@pace           STA VSYNC               ; enable screen to flip
@vsync          LDA VSYNC
                BNE @vsync              ; and wait for it to occur
                BIT VSYNC2
                BMI @ok
                STA VSYNC2              ; fast flip, gauge again for next frame
                ;
@ok     LDA VIC+$02
                EOR #$80
                STA VIC+$02             ; re-direct VIC to other screen buffer
                INC VCOUNT
                LDA SCRNPAGE
                STA PENDING             ; make active screen as pending
                EOR #$02
                STA SCRNPAGE    ; maintain VIC active video page
                ;
                ; PHASE IV:
                ; - write the same PLAYFIELD updates to new PENDING frame
                ;
                LDA #%00100000  ; clean PLAYFIELD dirt as part of this update
                JSR SSSCOMMIT
                ;
                PLA                             ;-- restore X (column)
                TAX
                PLA                             ;-- restore Y (row)
                TAY
                JSR SSSPLOT
                ;
@pause  ;JSR STOPKEY
                ;BNE @fini
                ;LDA #$00               ; clear skipped frame count
                ;STA VSYNC2
                ;STA VCOUNT
        ;       === write any custom PAUSE or RESET code here ===
                ;PLA
                ;PLA
                ;.global RESTART
                ;JMP RESTART            ; label speaks for itself
                ;
@fini   RTS


;*********************************************************************
; Software Sprite Stack COMMIT CHANGES TO PENDING FRAME BUFFER
;
; pass Accumulator with the PLAYFIELD bit(s) set for cleaning:
; - bit 7 for video #1
; - bit 6 for video #2
; - bit 5 for playfield
; - bit 4 for topfield
;
; SPRITE dirt for the PENDING frame is always cleaned.
;
; This is used by SSSFLIP and NOT normally called by user programs.
;
DIRTYMASK       = $00
CLEANER         = $01           ; bit 7=VIDEO1, 6=VIDEO2, 5=PLAYFIELD, 4=STATIC
;
SSSCOMMIT 
                LDX PENDING
                CPX #>VICFRAME1 ; will flip to video #1 next?
                BNE @scrn2
                ORA #%10000000  ; erase old sprites from video #1
                BNE @cont
@scrn2          ORA #%01000000  ; erase old sprites from video #2
@cont           TAY
                EOR #$FF        ; reverse check to clean
                STA CLEANER
                TYA
                ORA #%00100000  ; but always look for PLAYFIELD dirt
                STA DIRTYMASK
                TXA
                STA VECTOR2+1
                ORA #$84        ; and its COLOR
                STA VECTOR3+1
                LDX #$00
                STX VECTOR2
                STX VECTOR3
                LDY #$00
                JSR SSSPLOT     ; home "cursor"
                ;
                LDX #$00
@forx           LDA DIRTYLINE2,X
                STA ACOPY
                LDA DIRTYLINE,X ; dirty start column for this row
                CMP ACOPY
                BCS @nextx      ; reached end of this line?
                LDY PLAYCOLS
                STY NEWDIRT
                STY DIRTYLINE,X ; reset this line's seek for next commit
                LDY #$00
                STY DIRTYLINE2,X ; reset this line's end for next commit
                TAY              ; but start this commit from this column
@fory           LDA (COLORLINE),Y
                AND #%11100000  ; is this cell dirty for ANY update?
                BEQ @nexty           ; no, skip it
                STY DIRTYLINE2,X ; new ending column for this line
                INC DIRTYLINE2,X ; new ending column for this line
                CPY NEWDIRT
                BCS @more
                STY NEWDIRT
                STY DIRTYLINE,X ; new starting column for this line
@more           AND DIRTYMASK   ; is this cell dirty for THIS update?
                BEQ @nexty           ; no, skip it
                LDA (COLORLINE),Y
                STA (VECTOR3),Y ; update color cell
                AND CLEANER      ; remove this dirt from this cell
                STA (COLORLINE),Y
                LDA (SCRNLINE),Y
                STA (VECTOR2),Y ; update video cell
@nexty          INY
                CPY ACOPY
                BCC @fory
@nextx          LDA SCRNLINE
                CLC
                ADC PLAYCOLS
                BCC @cc
                INC SCRNLINE+1
                INC COLORLINE+1
                INC VECTOR2+1
                INC VECTOR3+1
@cc             STA SCRNLINE
                STA COLORLINE
                STA VECTOR2
                STA VECTOR3
                INX
                CPX PLAYROWS
                BCC @forx
                ;
                LDA #%10000000
                LDX PENDING
                CPX #>VICFRAME2 ; will flip to video #2 next?
                BNE @scrn1
                LDA #%01000000
@scrn1          STA NEWDIRT
                RTS

                
;*********************************************************************
; Software Sprite Stack CREATE A NEW SPRITE IN THE LIST
;
; pass Accumulator with the SPRITEDEF value (see HEADER)
; pass Y with the SPRITEH value (1-16)
; pass sssSPRITEDDEF2 with correct extended geometry 
; returns X with sprite index #0 thru SPRITEMAX-1
;
SSSCREATEEX 
                
                ; we are expecting a valid value in sssSPRITEDDEF2
                JMP DOSSSCREATE

SSSCREATE 

                PHA
                LDA #0
                STA sssSPRITEDDEF2
                PLA

DOSSSCREATE             
                
                STY YCOPY
                LDX SPRITES
                CPX #SPRITEMAX
                BCC @cont
                RTS       ; sorry, increase SPRITEMAX and re-compile
@cont 

                ; save supplied sprite def...
                STA SPRITEDEF,X
                
                ; Set default z-order to ordinal of the sprite...
                TXA
                STA SSSZORDR,X

                ; Check sprites < 24px high
                LDA sssSPRITEDDEF2
                BEQ @defaultSPRITEDDEF2
                AND #$03
                CMP #$03
                BNE @definitionOK
                BRK     ; invalid param
@definitionOK 
                LDA sssSPRITEDDEF2
                STA SPRITEDEF2,X
                JMP @suppliedSPRITEDDEF2

@defaultSPRITEDDEF2 

                LDA SPRITEDEF,X
                AND #%00001110
                ASL
                STA SPRITEDEF2,X
                LDA SPRITEDEF,X
                AND #%00000001
                ORA SPRITEDEF2,X
                STA SPRITEDEF2,X

@suppliedSPRITEDDEF2 
                
                CPX #$00
                BNE @append             ; >1 sprite
                ;
                ; this is the first sprite in the list ...
                LDA #<SSSBUF
                LDY #>SSSBUF
                STA SPRITEBUFL
                STY SPRITEBUFH
                LDY #$20                ; start at top of custom character
                STX SPRITEC1L
                STY SPRITEC1H
                STX SPRITEC2L
                STY SPRITEC2H
                BNE @compute
@append 
                ; copy prior sprite vectors
                LDA SPRITEBUFL-1,X
                STA SPRITEBUFL,X
                LDA SPRITEBUFH-1,X
                STA SPRITEBUFH,X
                LDA SPRITEC1L-1,X
                STA SPRITEC2L,X
                LDA SPRITEC1H-1,X
                STA SPRITEC1H,X
                STA SPRITEC2H,X
ifdef SPRITEDEF4
                ; repeating sprite?
                LDA SPRITEDEF,X
                AND #SPRITEDEF4
                BNE @same
endif

                ; no, allocate new image and character buffers
                LDA SPRITEDEF2-1,X
                AND #$1F
                TAY
                LDA sssALLOC,Y
ifdef DEBUG_CHECKS
                BNE @dbOK1
                byte $02   
@dbOK1 
endif
                CLC
                ADC SPRITEBUFL,X
                BCC @cc
                INC SPRITEBUFH,X
@cc             STA SPRITEBUFL,X
@compute 
                LDA SPRITEDEF2,X
                AND #$1F
                TAY
                LDA sssALLOC,Y
ifdef DEBUG_CHECKS
                BNE @dbOK2
                byte $02    
@dbOK2 
endif
                STA sssBYTES
                ; vector#2 into custom char set
                LDA SPRITEC2L,X
                SEC
                SBC sssBYTES    ; account for its entire buffer size
                BCS @cc1
                DEC SPRITEC2H,X
                DEC SPRITEC1H,X
@cc1            STA SPRITEC2L,X
                STA SPRITEC1L,X
                ; vector#1 into custom char set
                SEC
                SBC sssBYTES    ; account for its entire buffer size
                BCS @cc2
                DEC SPRITEC1H,X
@cc2            STA SPRITEC1L,X
                JMP @new
                ; keep repeating sprite pointing to same custom characters
@same           LDA SPRITEC1L-1,X
                STA SPRITEC1L,X
                LDA SPRITEC2L-1,X
                STA SPRITEC2L,X
                LDA SPRITEC2H-1,X
                STA SPRITEC2H,X
@new            LDA YCOPY
                STA SPRITEH,X
ifdef SPRITEDEF6
                LDA #SSSNULL    ; init with nothing in contact
                STA SPRITEBACK,X
                LDA #$00
                STA SPRITECX,X  ; init collision X-coord
                STA SPRITECY,X  ; init collision Y-coord
endif
                LDA #$00
                STA SPRITEX,X   ; sprite is not in visible area to start
                STA SPRITEY,X
                STA SPRITEZ,X   ; all flags off
                LDX SPRITES     ; return this new sprite # as initialized
                STX sssNUM
                INC SPRITES     ; account for the new sprite allocated
                RTS


;*********************************************************************
; Software Sprite Stack SELECT A SPRITE TO MANIPULATE
;
; pass X index with the SPRITES number (0 - <SPRITEMAX)
; preset values for sssNUM, sssX, sssY, sssBYTES, sssNEXT
;

SSSUSE 
                CPX SPRITES
                BCC @cont
                BRK              ; debugging is in your future
@cont           STX sssNUM
                LDA SPRITEDEF2,X
                AND #$1F
                TAY             ; save for later
                AND #%00001000  ; Y-float enabled?
                BEQ @y          ; branch if fixed
                LDA SPRITEY,X   
                AND #$07        ; mask the cell offset
                BNE @y          ; branch if not on cell boundary
                TYA             ; recover spritedef2
                AND #%00010111  ; mask out Y float
                TAY             ; and save
@y              TYA             ; recover 
                AND #%00010000  ; X-float enabled?
                BEQ @x          ; branch if X is fixed
                LDA SPRITEX,X   ; get the X value
                AND #$07        ; mask the cell offset
                BNE @x          ; branch  if not on cell boundary
                TYA             ; recover processed spritdef2
                AND #%00001111  ; mask out X float
                TAY             ; set as index to cell count 
@x              ;
                LDA sssALLOC,Y  ; get the number of character cells needed to plot sprite

ifdef DEBUG_CHECKS
                BNE @dbOK3
                byte $02    
@dbOK3 
endif
                STA sssBYTES    ; and store
                ;
                LDA sssROWS,Y   ; get 
                STA sssY
                ASL
                ASL
                ASL             ; x8
                STA sssNEXT
                ;
                LDA sssCOLS,Y   ;
                STA sssX
@fini           RTS


;*********************************************************************
; Software Sprite Stack LOAD SPRITE BUFFER WITH NEW IMAGE DATA
;
; SSSUSE must be called prior to point to sprite register entry
; pass Accumulator with this sprite's color
; pass X,Y as the source image pointer
;
SSSANIM 
                STX sssDX         ;low address of animation data
                LDX sssNUM        ;sprite ID
                STA SPRITECOL,X   ;store sprite colour
                LDA sssDX
                STA SPRITEIMGL,X  ;store low address
                TYA
                STA SPRITEIMGH,X  ;store high address
                BNE SSSTOUCH


;*********************************************************************
; Software Sprite Stack MOVE A SPRITE TO ABSOLUTE X,Y COORDINATES
;
; SSSUSE must be called prior to point to sprite register entry
; pass X,Y with the sprite coordinates
;
SSSMOVEXY 
                TXA
                LDX sssNUM    
                STA SPRITEX,X
                TYA
                STA SPRITEY,X


;*********************************************************************
; Software Sprite Stack FLAG A SPRITE FOR RENDERING
;
; Pass X with the sprite#
;
SSSTOUCH 
                LDA SPRITEZ,X
                AND #%11        ; reset flags, except fast copy + custom char
                ORA #%11000000  ; force sprite to do make + copy/merge
                STA SPRITEZ,X
                RTS


;*********************************************************************
; Software Sprite Stack FLAG ALL SPRITES FOR RENDERING
;
SSSREFRESH 
                LDX #$00
@loop           CPX SPRITES
                BCS @fini
                JSR SSSTOUCH
                INX
                BNE @loop
@fini           RTS



;*********************************************************************
; Software Sprite Stack UPDATE PENDING FRAME BUFFER WITH SPRITES
;
; This is part of the SSSFLIP operation and is NOT expected to be
; called by user programs.
;
SSSUPDATE 
                LDX #0
                STX sssIDX   ;mod: sssnum in original
@do 
                CPX SPRITES
                BCC @cc
                RTS
@cc             
                ; Mod draw sprites in Z_ORDER
                LDA SSSZORDR,X
                STA sssNUM
                TAX
                
                ; Previous code
                LDA SPRITEDEF,X

ifdef SPRITEDEF4        ; repeating sprite?
                AND #SPRITEDEF4
                BEQ @own        ;skip this sprite is independent
                LDA SPRITEZ-1,X
                STA SPRITEZ,X   ; copy Z-flags for display & results
                JMP @matrix
else
                ASL
                BCC @redraw             ; sprite is disabled
                LDA SPRITEX,X
                BEQ @redraw             ; 0 = outside left border visible range
                CMP SSSCLIPX
                BCS @redraw             ; >= outside right border visible range
                LDA SPRITEY,X
                BEQ @redraw             ; 0 = outside top border visible range
                CMP SSSCLIPY
                BCS @redraw             ; >= outside bottom border visible range
endif

                ; preset sprite image buffer:
                ; VECTOR1 = pointer to top-left within sprite matrix
                ; VECTOR2&3 = pointer to adjacent chars, as necessary
@own    
                JSR SSSUSE
                LDA SPRITEBUFH,X
                STA VECTOR1+1
                STA VECTOR2+1

ifdef SPRITEWIDE
                STA VECTOR3+1
endif

                LDA SPRITEBUFL,X
                STA VECTOR1
                CLC
                ADC sssNEXT
                BCC @cc1
                INC VECTOR2+1

ifdef SPRITEWIDE
                INC VECTOR3+1
endif

@cc1            STA VECTOR2

ifdef SPRITEWIDE
                CLC
                ADC sssNEXT
                BCC @cc2
                INC VECTOR3+1
@cc2            STA VECTOR3
endif

ifdef SPRITEDEF5   ;ghost display allowed
                LDY #$11                ; ORA opcode
                LDA SPRITEDEF,X
                AND #SPRITEDEF5         ; ghost image specified
                BEQ @bit
                LDY #$51                ; EOR opcode
@bit            TYA
                STA @OP1
ifdef SPRITEDEF6 ;Collision detection allowed
                STA @OP2
                EOR #$40                ; swap opcode
                STA @OP3
endif
endif

                ; branch on make control flags
                LDA SPRITEZ,X
                ASL
                PHA
                BCC @copy1              ; $80 - (re)make buffered image
                JSR @Make
@copy1          PLA
                ASL
                BCC @matrix             ; $40 - copy buffered image
                JSR @Copy
@matrix 
                JSR @Display    ; display sprite matrix
                JMP @loop
@redraw 
                LDA SPRITEZ,X
                AND #%11
                ORA #%11110000  ; make + copy/merge + null bg + clipped fg
                STA SPRITEZ,X
@loop 
                INC sssIDX
                LDX sssIDX
                JMP @do
                ;
                ; INIT PHASE
                ; ----------
                ; (re)make this sprite's image buffer
@Make   ;
                LDA SPRITEH,X
                STA sssXFER             ; sprite image raster count
                LDA SPRITEY,X
                AND #$07
                CLC
                ADC sssXFER
                STA sssDY       ; 1st raster below image
                ;
                ; VECTORBG = pointer to your compact source image
                LDA SPRITEIMGL,X
                STA VECTORBG
                LDA SPRITEIMGH,X
                STA VECTORBG+1
                LDA sssNEXT
                STA sssLINENUM  ; this many raster lines to copy
                CMP sssDY
                BCS @Mloop              ; fits within height of sprite
                LDA sssDY
                SEC
                SBC sssLINENUM
                STA ACOPY               ; compute how many rasters to clip
                LDA sssXFER
                SBC ACOPY
                STA sssXFER             ; clip image within sprite height
@Mloop 
                LDA #$00                ; erase raster registers
                STA sssROR1
                STA sssROR2

ifdef SPRITEWIDE
                STA sssROR3
endif

                DEC sssLINENUM
                LDY sssLINENUM
                LDA sssXFER
                BEQ @Mcopy              ; no more rasters to copy - zero them
                CPY sssDY
                BCS @Mcopy              ; below sprite image - zero this raster
@Mxfer 
                DEC sssXFER             ; copying sprite image rasters
                LDY sssXFER
                LDA (VECTORBG),Y
                STA sssROR1

                LDA SPRITEDEF2,X     ;mod using def2 
                AND #%00000100
                BEQ @Mxfer2             ; 16w ?
                
                ; set height of sprites..
                LDA SPRITEDEF2,X
                AND #$03
                TAY
                LDA sssHEIGHTS,Y  
                CLC
                ADC sssXFER
                TAY
                LDA (VECTORBG),Y
                STA sssROR2     ; load adjacent register
@Mxfer2 
                LDA SPRITEX,X
                AND #$07
                BEQ @Mcopy
                STA sssDX
@Mx2            LSR sssROR1
                ROR sssROR2             ; shift into image overflow register #1

ifdef SPRITEWIDE
                ROR sssROR3             ; shift into image overflow register #2
endif

                DEC sssDX
                BNE @Mx2
@Mcopy 
                LDX #$01
                LDY sssLINENUM
                LDA sssROR1
                STA (VECTOR1),Y
                CPX sssX
                BEQ @Mnext
                LDA sssROR2             ; write image overflow register #1
                STA (VECTOR2),Y

ifdef SPRITEWIDE
                INX
                CPX sssX
                BEQ @Mnext
                LDA sssROR3             ; write image overflow register #2
                STA (VECTOR3),Y
endif

@Mnext 
                LDX sssNUM
                LDA sssLINENUM
                BEQ @Mfini
                JMP @Mloop
@Mfini          RTS
                ;
                ; PHASE II
                ; --------
                ; copy/merge buffered image with background into
                ; sprite character matrix
@Copy   ;
                LDA SPRITEZ,X
                AND #%11
                EOR #%1                 ; flip to other character set
                STA SPRITEZ,X
                AND #%1
                BNE @Cfb2
                LDA SPRITEC1L,X
                STA VECTORFG
                LDA SPRITEC1H,X
                STA VECTORFG+1
                BNE @Ccopy
@Cfb2           LDA SPRITEC2L,X
                STA VECTORFG
                LDA SPRITEC2H,X
                STA VECTORFG+1
@Ccopy 
                LDA SPRITEX,X
                STA sssDX
                LDA sssNEXT
                STA sssLINE
                LDY #0
                STY sssCHAR
                STY sssLINENUM
                STY ACOPY
@Cdocol 
                LDX sssNUM
                LDA SPRITEY,X
                STA sssDY
@Cbgimage 
                LDX sssNUM
                LDA SPRITEZ,X
                AND #%10
                BNE @Cjmp               ; fast copy?
                LDX sssDX
                LDY sssDY
                JSR SSSREAD
                CMP #SSSNULL
                BNE @Cmore
@Cjmp           JMP @Ccpfast
@Cmore 
                LDX sssNUM
                LDA CRSRCOLOR
                AND #%10000
                BEQ @Ccont              ; static cell?
                LDA SPRITEZ,X
                ORA #%10000             ; flag that sprite's foreground is clipped
                STA SPRITEZ,X
                
                ; Below shortcut commented out as doesn't work with 24 high sprites....
                ;JMP @Cnextrow  ; don't bother merging with a background
@Ccont 

ifdef SPRITEDEF4  ;repeat sprites allowed
                LDA SPRITEDEF,X
                AND #SPRITEDEF4
                BNE @Cjmp      ;repeat not specified
endif

                LDA CRSRCHAR
                JSR SSSIMAGE
                STX VECTORBG
                STY VECTORBG+1
                LDX #0
                STX sssXFER
                INC sssCHAR     ; flag that there is something behind this sprite

                LDY sssNUM
                LDA SPRITECOL,Y ; save MC status of sprite
                AND #8
                STA sssMC   
                
ifdef SPRITEDEF6   ;sprite collisions allowed
                LDY sssNUM
                LDA SPRITEZ,Y
                AND #%1000
                BNE @Ccploop    ; already a collision?
                LDA SPRITEDEF,Y
                AND #SPRITEDEF6
                BNE @Ccploopx   ; collision detection enabled?
endif

@Ccploop 
                LDA sssMC
                BEQ @normal
@McMerge 
                ;Multicolour data merge
                STY YCOPY
                LDY sssXFER       ; from ...
                LDA (VECTORBG),Y
                BEQ @afterNormal
                STA zp1           ; background pixels
                LDY sssLINENUM    ; to ...
                LDA (VECTOR1),Y 
                STA zp2           ; foreground pixels     
        
               ; Save the FG mask
                TAY 
                LDA COLORMAPINDEX,Y
                PHA               ; save for later....

                AND zp1           ;and with Background
                STA zp3           ; store to OUTPUT BUFFER
                
                PLA
                EOR #$FF          ; invert it
                AND zp2           ; and with foregrond
                ORA zp3           ; merge with background
                
@saveResult 

                LDY sssLINENUM     ; to ...
                STA (VECTORFG),Y
                LDY YCOPY
                jmp @Ccpnxt
@normal 
                ;Normal colour data merge
                LDY sssXFER        ; from ...
                LDA (VECTORBG),Y
@afterNormal 
                LDY sssLINENUM     ; to ...
@OP1            ORA (VECTOR1),Y    ; opcode modification #1: ORA / EOR
                STA (VECTORFG),Y
@Ccpnxt 
                INC sssLINENUM
                INC sssXFER
                INX
                CPX #8
                BNE @Ccploop
                BEQ @Cnextrow

ifdef SPRITEDEF6  ;Collision detection allowed
@Ccploopx 
                LDY sssXFER        ; from ...
                LDA (VECTORBG),Y
                PHA                ;++
                LDY sssLINENUM    ; to ...
@OP2            ORA (VECTOR1),Y   ; opcode modification #2: ORA / EOR
                STA (VECTORFG),Y
                PLA               ;--
@OP3            EOR (VECTOR1),Y  ; opcode modification #3: ORA / EOR
                CMP (VECTORFG),Y
                BEQ @Cnohit      ; any overlapping pixel(s)?
                LDY sssNUM
                LDA sssDX
                STA SPRITECX,Y   ; save X sprite coord of what was hit
                LDA sssDY
                STA SPRITECY,Y   ; save Y sprite coord of what was hit
                LDA CRSRCHAR     ; save character code of what was hit
                STA SPRITEBACK,Y
                LDA SPRITEZ,Y
                ORA #%1000       ; sprite-pixel collision with non-static cell
                STA SPRITEZ,Y
                BNE @Ccpnxt      ; resume normal copy/merge operation
@Cnohit 
                INC sssLINENUM
                INC sssXFER
                INX
                CPX #8
                BNE @Ccploopx
                BEQ @Cnextrow
endif

                ; faster copy, because there is no backgound to merge with ... 
@Ccpfast                                
                LDX #8
                LDY sssLINENUM
@Ccploop2 
                LDA (VECTOR1),Y
                STA (VECTORFG),Y
                INY
                DEX
                BNE @Ccploop2
                STY sssLINENUM
@Cnextrow 
                LDA sssDY
                CLC
                ADC #8
                STA sssDY
                LDY sssLINENUM
                CPY sssLINE
                BEQ @Cnextcol
                JMP @Cbgimage
@Cnextcol 
                LDA ACOPY
                CLC
                ADC sssNEXT
                STA ACOPY
                STA sssLINENUM
                LDA sssLINE
                CLC
                ADC sssNEXT
                STA sssLINE
                LDA sssDX
                CLC
                ADC #$08
                STA sssDX
                DEC sssX
                BEQ @Cdone
                JMP @Cdocol
@Cdone 
                LDA sssCHAR
                BNE @Cfini
                LDX sssNUM      ; all null background, if no new changes occur
                LDA SPRITEZ,X   ; don't do a merge on next flip either
                ORA #$20        ; enable this sprite to be re-used as-is
                STA SPRITEZ,X   ; on next frame flip
@Cfini          RTS
                ;
                ; PHASE III
                ; ---------
                ; display sprite character matrix
                ; by row, then by column
@Display 
                LDX sssNUM

ifdef SPRITEDEF4  ;repeat sprites are permitted
                LDA SPRITEDEF,X
                ASL
                BCC @Cfini        ; sprite is disabled
endif

                JSR SSSUSE        ;calculate sprite character requirement at this position
                DEC sssY          ;modify the calculated row and column cell counts
                DEC sssX
                LDA SPRITEX,X
                STA sssDX         ; get the current Xpos
                LDA SPRITECOL,X   ; get the current colour
                STA COLORCODE     ; save as the playfield character colour

ifdef SPRITEDEF5  ;ghost sprites are permitted
                LDA SPRITEDEF,X
                AND #SPRITEDEF5 ; ghost mode specified
                BEQ @Dok        ; branch if in normal display mode
                LDA COLORCODE
                ORA #$80
                STA COLORCODE   ; flag to keep PLAYFIELD colored cells
endif

@Dok            LDA SPRITEZ,X   ;display control
                AND #%1         ; last rendered true: c2 last used
                BNE @Dfb2       ; branch if C2
                LDA SPRITEC1L,X ; set up the pointers for the Foreground
                STA VECTORFG
                LDY SPRITEC1H,X 
                BNE @Dchar
@Dfb2           LDA SPRITEC2L,X
                STA VECTORFG
                LDY SPRITEC2H,X 
@Dchar          STY VECTORFG+1  
                TYA
                SEC            
                SBC #UDCSTART  ; Start Page for custom chars
                TAY            ; index into multiply by 32 table
                LDA ASL5,Y     ; Y * 32

                STA sssCHAR
                LDA VECTORFG
                LSR
                LSR
                LSR              ; /8
                ADC sssCHAR
                STA sssCHAR      ; start with this custom character
@Dcol 
                LDA sssY         ; use sprite character row
                STA sssLINENUM   ; prepare for character write
                LDX sssNUM
                LDA SPRITEY,X    ; get the sprite Y position
                STA sssDY        ; custom character row (0, 1?, 2?)
@Drow 
                LDA sssCHAR
                LDX sssDX
                LDY sssDY
                JSR SSSWRITE    ; display it
@Dskip 
                INC sssCHAR             ; account for it, even if it is not displayed
                LDA sssLINENUM
                BEQ @Dnrow
                DEC sssLINENUM
                LDA sssDY
                CLC
                ADC #8
                STA sssDY               ; next Y-pixel
                JMP @Drow
@Dnrow 
                LDA sssDX
                CLC
                ADC #8
                STA sssDX
                LDA sssX
                BEQ @Dncol
                DEC sssX
                JMP @Dcol
@Dncol 
                RTS


;*********************************************************************
; Software Sprite Stack GET IMAGE ADDRESS FROM A CHARACTER
;
; This is used by SSSUPDATE and NOT normally called by user programs.
;
; Pass A with the character code.
; returns X/Y as a pointer to its image source.
;

SSSIMAGE
               TAY                ;save id
               ASL
               ASL
               ASL                 ; x8
               TAX                 ; save image low byte
               TYA                 ; recover id
               ROL                 ; set carry bit if char >7f (ROM)
               LDY  #$00           ; point to custom chars
               BCC  @cont          ; branch if udc
ifdef SSSROMCHARS
               INY                 ; point to ROM chars
endif
@cont          ROL
               ROL
               ROL                 ; rotate top 4 bits into low nibble
               AND  #%00000011     ; mask bottom 2 completes id * 8
               ORA  @vic,Y         ; prepend page pointer
               TAY                 ; save image high byte
               RTS
                                   ; VIC custom or ROM characters
@vic    byte   UDCSTART, $80

   
;*********************************************************************
; Software Sprite Stack READ FROM PENDING FRAME BUFFER
;
; This is used by SSSUPDATE and NOT normally called by user programs.
;
; Pass X/Y with a sprite pixel coordinate.
; CURSOR is re-plotted to this location.
; returns Accumulator with character code from PENDING frame buffer,
; or a SPACE if the coordinate is outside the screen borders.
; returns X/Y cell coordinates.
;
SSSREAD 
                LDA #SSSNULL    ; default to an empty background
                CPX #$10
                BCC @fini               ; hidden by left border
                CPX SSSCLIPX
                BCS @fini               ; hidden by right border
                CPY #$10
                BCC @fini               ; above top border
                CPY SSSCLIPY
                BCS @fini               ; below bottom border
                ;
                LDA PENDING
                STA SCRNLINE+1
                LDA #>PLAYCOLOR
                STA COLORLINE+1
                ;
                TYA
                SEC
                SBC #$10
                TAY
                LDA LSR3,Y   ;divide Y by 8
                TAY
                STY CRSRROW
                ;
                TXA
                SEC
                SBC #$10
                TAX                         
                LDA LSR3,X   ;divide X by 8
                STA CRSRCOL
                ;
@ok             TYA
                ASL
                TAY
                LDA sss+1,Y
                BEQ @top
                INC SCRNLINE+1
                INC COLORLINE+1
@top            LDA sss,Y
                STA SCRNLINE
                STA COLORLINE
                LDY CRSRCOL
                LDA (COLORLINE),Y       ; read from pending color buffer
                STA CRSRCOLOR
                LDA (SCRNLINE),Y        ; read from pending video buffer
                STA CRSRCHAR
                LDY CRSRROW
@fini           RTS


;*********************************************************************
; Software Sprite Stack WRITE TO PENDING FRAME BUFFER
;
; This is used by SSSUPDATE and NOT normally called by user programs.
;
; Pass X/Y with a sprite pixel coordinate.
; Pass A with SPRITE character code to write to PENDING screen buffer.
; COLORCODE is used to fill same space with that value.
;
; Write does not occur if X/Y lie outside the screen borders.
;
SSSWRITE 
                CPX #$10
                BCC @fini               ; hidden by left border
                CPX SSSCLIPX
                BCS @fini               ; hidden by right border
                CPY #$10
                BCC @fini               ; above top border
                CPY SSSCLIPY
                BCS @fini               ; below bottom border
                PHA                     ;++
                JSR SSSPEEKXY
                LDA CRSRCOLOR
                AND #$10                ; static cell?
                BEQ @ok
                PLA                     ; yes, don't overwrite it!
                RTS
                ;
@ok             LDX CRSRCOL
                LDY CRSRROW
                JSR SSSPLOTS
                PLA                     ;--
                JSR SSSPOKE
@fini           RTS


;*********************************************************************
; Software Sprite Stack PROTECTED WRITE TO PENDING FRAME BUFFER
;
; Pass X/Y with a screen cell coordinate.
; Pass A with character code to write to PENDING screen buffer.
; COLORCODE is used to fill same space with that value.
;
SSSCELL 
                PHA               ;++
                LDA NEWDIRT
                STA $01
                LDA #$10
                STA NEWDIRT
                JSR SSSPLOTS
                PLA               ;--
                JSR SSSPOKE
                LDA $01
                STA NEWDIRT
                RTS



;*********************************************************************
; Software Sprite Stack LOAD BACKGROUND
;
; Address of data follows call: word 0 address of char map, word 1 address of colour map
; Value in Ac $FF use map, otherwise use colour in Ac
; Data is written to the PLAYFIELD AND PLAYCOLOR BUFFERS
;
; These changes will go into effect only after a call to SSSFLIP
;

SSSLOADBG       sta colAction  
                pla
                sta VECTORFG
                pla
                sta VECTORFG+1
                ldy #1
                lda (VECTORFG),y
                sta zp1
                iny
                lda (VECTORFG),y
                sta zp1+1
                bit colAction
                bpl @f1
                iny
                lda (VECTORFG),y
                sta zp3
                iny
                lda (VECTORFG),y
                sta zp3+1
@f1             tya             ;calculate new RTS address
                clc
                adc VECTORFG
                sta VECTORFG
                lda #0
                adc VECTORFG+1
                pha             ;push back on stack
                lda VECTORFG
                pha
                LDX #$00
                LDY #$00
                JSR SSSPLOT             ; home "cursor"
                ;assess screen size
                lda VIC + 2 ;number of cols
                and #$7f
                tax
                lda VIC + 3 ;number of rows
                and #$7e
                lsr 
                tay               
                lda #0
                sta ssscells+1
                stx ssscells
@b2             clc
                adc ssscells
                bcc @f2a
                inc ssscells+1
@f2a            dey 
                bne @b2
                sta ssscells  ;number of cells

                ldy #0
                sty ssscount 
                dec ssscount  ;initialise to -1
@b1             sty sssty
                lda colAction    
                bpl @f2
                lda (zp3),y
@f2             sta COLORCODE
                lda (zp1),y
                JSR SSSPRINT  ;put character and colour into the Pending write data
                dec ssscells
                bne @f3
                dec ssscells+1
                bmi @end
@f3             ldy sssty
                iny
                bne @b1
                inc zp1+1
                inc zp3+1
                inc ssscount
                beq @b1          ;repeat ONCE
@end            RTS
               
;local data
colaction byte 0
ssscells  word 0 
ssscount  byte 0
sssty     byte 0   

; LOOKUP TABLES
                        ;              H x W [0 = configuration not supported]
sssALLOC        ;       8x8, 16x8, 24x8, [32x8], 8x16, 16x16, 24x16, [32x 16] 
                byte     8,16,24,0,16,32,48,0    ; fixed:            1,2,3,n/a,2,4,6,n/a
                byte    16,24,32,0,32,48,64,0   ; float Y:          2,3,4,n/a,4,6,8,n/a
                byte    16,32,48,0,24,48,72,0   ; float X:          2,4,6,n/a,3,6,9,n/a
                byte    32,48,64,0,48,72,96,0   ; float both:       4,6,8,n/a,9,6,12,n/a

sssROWS         byte    1,2,3,0,1,2,3,0 ; fixed
                byte    2,3,4,0,2,3,4,0 ; float Y
                byte    1,2,3,0,1,2,3,0 ; float X
                byte    2,3,4,0,2,3,4,0 ; float XY

sssCOLS         byte    1,1,1,0,2,2,2,0 ; fixed
                byte    1,1,1,0,2,2,2,0 ;  float Y
                byte    2,2,2,0,3,3,3,0 ;  float X
                byte    2,2,2,0,3,3,3,0 ;  float XY
   

ASL5    ;look up for multiply integer in range 0-8 by 32
        byte 0,32,64,96,128,160,192,224,0  
              
        
LSR3     ;look up for integer VALUE/8 in the range 0-255
         bytes 0*8  ;0,0,0,0,0,0,0
         bytes 1*8  ;1,1,1,1,1,1,1
         bytes 2*8
         bytes 3*8
         bytes 4*8
         bytes 5*8
         bytes 6*8
         bytes 7*8
         bytes 8*8
         bytes 9*8
         bytes 10*8 
         bytes 11*8
         bytes 12*8
         bytes 13*8
         bytes 14*8
         bytes 15*8
         bytes 16*8
         bytes 17*8
         bytes 18*8
         bytes 19*8
         bytes 20*8
         bytes 21*8 
         bytes 22*8
         bytes 23*8
         bytes 24*8
         bytes 25*8
         bytes 26*8
         bytes 27*8
         bytes 28*8
         bytes 29*8
         bytes 30*8
         bytes 31*8
         bytes 32*8 



; Color map table 
COLORMAPINDEX 
Index0  byte %11111111; input= %00000000
Index1  byte %11111100; input= %00000001
Index2  byte %11111100; input= %00000010
Index3  byte %11111100; input= %00000011
Index4  byte %11110011; input= %00000100
Index5  byte %11110000; input= %00000101
Index6  byte %11110000; input= %00000110
Index7  byte %11110000; input= %00000111
Index8  byte %11110011; input= %00001000
Index9  byte %11110000; input= %00001001
Index10  byte %11110000; input= %00001010
Index11  byte %11110000; input= %00001011
Index12  byte %11110011; input= %00001100
Index13  byte %11110000; input= %00001101
Index14  byte %11110000; input= %00001110
Index15  byte %11110000; input= %00001111
Index16  byte %11001111; input= %00010000
Index17  byte %11001100; input= %00010001
Index18  byte %11001100; input= %00010010
Index19  byte %11001100; input= %00010011
Index20  byte %11000011; input= %00010100
Index21  byte %11000000; input= %00010101
Index22  byte %11000000; input= %00010110
Index23  byte %11000000; input= %00010111
Index24  byte %11000011; input= %00011000
Index25  byte %11000000; input= %00011001
Index26  byte %11000000; input= %00011010
Index27  byte %11000000; input= %00011011
Index28  byte %11000011; input= %00011100
Index29  byte %11000000; input= %00011101
Index30  byte %11000000; input= %00011110
Index31  byte %11000000; input= %00011111
Index32  byte %11001111; input= %00100000
Index33  byte %11001100; input= %00100001
Index34  byte %11001100; input= %00100010
Index35  byte %11001100; input= %00100011
Index36  byte %11000011; input= %00100100
Index37  byte %11000000; input= %00100101
Index38  byte %11000000; input= %00100110
Index39  byte %11000000; input= %00100111
Index40  byte %11000011; input= %00101000
Index41  byte %11000000; input= %00101001
Index42  byte %11000000; input= %00101010
Index43  byte %11000000; input= %00101011
Index44  byte %11000011; input= %00101100
Index45  byte %11000000; input= %00101101
Index46  byte %11000000; input= %00101110
Index47  byte %11000000; input= %00101111
Index48  byte %11001111; input= %00110000
Index49  byte %11001100; input= %00110001
Index50  byte %11001100; input= %00110010
Index51  byte %11001100; input= %00110011
Index52  byte %11000011; input= %00110100
Index53  byte %11000000; input= %00110101
Index54  byte %11000000; input= %00110110
Index55  byte %11000000; input= %00110111
Index56  byte %11000011; input= %00111000
Index57  byte %11000000; input= %00111001
Index58  byte %11000000; input= %00111010
Index59  byte %11000000; input= %00111011
Index60  byte %11000011; input= %00111100
Index61  byte %11000000; input= %00111101
Index62  byte %11000000; input= %00111110
Index63  byte %11000000; input= %00111111
Index64  byte %00111111; input= %01000000
Index65  byte %00111100; input= %01000001
Index66  byte %00111100; input= %01000010
Index67  byte %00111100; input= %01000011
Index68  byte %00110011; input= %01000100
Index69  byte %00110000; input= %01000101
Index70  byte %00110000; input= %01000110
Index71  byte %00110000; input= %01000111
Index72  byte %00110011; input= %01001000
Index73  byte %00110000; input= %01001001
Index74  byte %00110000; input= %01001010
Index75  byte %00110000; input= %01001011
Index76  byte %00110011; input= %01001100
Index77  byte %00110000; input= %01001101
Index78  byte %00110000; input= %01001110
Index79  byte %00110000; input= %01001111
Index80  byte %00001111; input= %01010000
Index81  byte %00001100; input= %01010001
Index82  byte %00001100; input= %01010010
Index83  byte %00001100; input= %01010011
Index84  byte %00000011; input= %01010100
Index85  byte %00000000; input= %01010101
Index86  byte %00000000; input= %01010110
Index87  byte %00000000; input= %01010111
Index88  byte %00000011; input= %01011000
Index89  byte %00000000; input= %01011001
Index90  byte %00000000; input= %01011010
Index91  byte %00000000; input= %01011011
Index92  byte %00000011; input= %01011100
Index93  byte %00000000; input= %01011101
Index94  byte %00000000; input= %01011110
Index95  byte %00000000; input= %01011111
Index96  byte %00001111; input= %01100000
Index97  byte %00001100; input= %01100001
Index98  byte %00001100; input= %01100010
Index99  byte %00001100; input= %01100011
Index100  byte %00000011; input= %01100100
Index101  byte %00000000; input= %01100101
Index102  byte %00000000; input= %01100110
Index103  byte %00000000; input= %01100111
Index104  byte %00000011; input= %01101000
Index105  byte %00000000; input= %01101001
Index106  byte %00000000; input= %01101010
Index107  byte %00000000; input= %01101011
Index108  byte %00000011; input= %01101100
Index109  byte %00000000; input= %01101101
Index110  byte %00000000; input= %01101110
Index111  byte %00000000; input= %01101111
Index112  byte %00001111; input= %01110000
Index113  byte %00001100; input= %01110001
Index114  byte %00001100; input= %01110010
Index115  byte %00001100; input= %01110011
Index116  byte %00000011; input= %01110100
Index117  byte %00000000; input= %01110101
Index118  byte %00000000; input= %01110110
Index119  byte %00000000; input= %01110111
Index120  byte %00000011; input= %01111000
Index121  byte %00000000; input= %01111001
Index122  byte %00000000; input= %01111010
Index123  byte %00000000; input= %01111011
Index124  byte %00000011; input= %01111100
Index125  byte %00000000; input= %01111101
Index126  byte %00000000; input= %01111110
Index127  byte %00000000; input= %01111111
Index128  byte %00111111; input= %10000000
Index129  byte %00111100; input= %10000001
Index130  byte %00111100; input= %10000010
Index131  byte %00111100; input= %10000011
Index132  byte %00110011; input= %10000100
Index133  byte %00110000; input= %10000101
Index134  byte %00110000; input= %10000110
Index135  byte %00110000; input= %10000111
Index136  byte %00110011; input= %10001000
Index137  byte %00110000; input= %10001001
Index138  byte %00110000; input= %10001010
Index139  byte %00110000; input= %10001011
Index140  byte %00110011; input= %10001100
Index141  byte %00110000; input= %10001101
Index142  byte %00110000; input= %10001110
Index143  byte %00110000; input= %10001111
Index144  byte %00001111; input= %10010000
Index145  byte %00001100; input= %10010001
Index146  byte %00001100; input= %10010010
Index147  byte %00001100; input= %10010011
Index148  byte %00000011; input= %10010100
Index149  byte %00000000; input= %10010101
Index150  byte %00000000; input= %10010110
Index151  byte %00000000; input= %10010111
Index152  byte %00000011; input= %10011000
Index153  byte %00000000; input= %10011001
Index154  byte %00000000; input= %10011010
Index155  byte %00000000; input= %10011011
Index156  byte %00000011; input= %10011100
Index157  byte %00000000; input= %10011101
Index158  byte %00000000; input= %10011110
Index159  byte %00000000; input= %10011111
Index160  byte %00001111; input= %10100000
Index161  byte %00001100; input= %10100001
Index162  byte %00001100; input= %10100010
Index163  byte %00001100; input= %10100011
Index164  byte %00000011; input= %10100100
Index165  byte %00000000; input= %10100101
Index166  byte %00000000; input= %10100110
Index167  byte %00000000; input= %10100111
Index168  byte %00000011; input= %10101000
Index169  byte %00000000; input= %10101001
Index170  byte %00000000; input= %10101010
Index171  byte %00000000; input= %10101011
Index172  byte %00000011; input= %10101100
Index173  byte %00000000; input= %10101101
Index174  byte %00000000; input= %10101110
Index175  byte %00000000; input= %10101111
Index176  byte %00001111; input= %10110000
Index177  byte %00001100; input= %10110001
Index178  byte %00001100; input= %10110010
Index179  byte %00001100; input= %10110011
Index180  byte %00000011; input= %10110100
Index181  byte %00000000; input= %10110101
Index182  byte %00000000; input= %10110110
Index183  byte %00000000; input= %10110111
Index184  byte %00000011; input= %10111000
Index185  byte %00000000; input= %10111001
Index186  byte %00000000; input= %10111010
Index187  byte %00000000; input= %10111011
Index188  byte %00000011; input= %10111100
Index189  byte %00000000; input= %10111101
Index190  byte %00000000; input= %10111110
Index191  byte %00000000; input= %10111111
Index192  byte %00111111; input= %11000000
Index193  byte %00111100; input= %11000001
Index194  byte %00111100; input= %11000010
Index195  byte %00111100; input= %11000011
Index196  byte %00110011; input= %11000100
Index197  byte %00110000; input= %11000101
Index198  byte %00110000; input= %11000110
Index199  byte %00110000; input= %11000111
Index200  byte %00110011; input= %11001000
Index201  byte %00110000; input= %11001001
Index202  byte %00110000; input= %11001010
Index203  byte %00110000; input= %11001011
Index204  byte %00110011; input= %11001100
Index205  byte %00110000; input= %11001101
Index206  byte %00110000; input= %11001110
Index207  byte %00110000; input= %11001111
Index208  byte %00001111; input= %11010000
Index209  byte %00001100; input= %11010001
Index210  byte %00001100; input= %11010010
Index211  byte %00001100; input= %11010011
Index212  byte %00000011; input= %11010100
Index213  byte %00000000; input= %11010101
Index214  byte %00000000; input= %11010110
Index215  byte %00000000; input= %11010111
Index216  byte %00000011; input= %11011000
Index217  byte %00000000; input= %11011001
Index218  byte %00000000; input= %11011010
Index219  byte %00000000; input= %11011011
Index220  byte %00000011; input= %11011100
Index221  byte %00000000; input= %11011101
Index222  byte %00000000; input= %11011110
Index223  byte %00000000; input= %11011111
Index224  byte %00001111; input= %11100000
Index225  byte %00001100; input= %11100001
Index226  byte %00001100; input= %11100010
Index227  byte %00001100; input= %11100011
Index228  byte %00000011; input= %11100100
Index229  byte %00000000; input= %11100101
Index230  byte %00000000; input= %11100110
Index231  byte %00000000; input= %11100111
Index232  byte %00000011; input= %11101000
Index233  byte %00000000; input= %11101001
Index234  byte %00000000; input= %11101010
Index235  byte %00000000; input= %11101011
Index236  byte %00000011; input= %11101100
Index237  byte %00000000; input= %11101101
Index238  byte %00000000; input= %11101110
Index239  byte %00000000; input= %11101111
Index240  byte %00001111; input= %11110000
Index241  byte %00001100; input= %11110001
Index242  byte %00001100; input= %11110010
Index243  byte %00001100; input= %11110011
Index244  byte %00000011; input= %11110100
Index245  byte %00000000; input= %11110101
Index246  byte %00000000; input= %11110110
Index247  byte %00000000; input= %11110111
Index248  byte %00000011; input= %11111000
Index249  byte %00000000; input= %11111001
Index250  byte %00000000; input= %11111010
Index251  byte %00000000; input= %11111011
Index252  byte %00000011; input= %11111100
Index253  byte %00000000; input= %11111101
Index254  byte %00000000; input= %11111110
Index255  byte %00000000; input= %11111111